/*

=head1 NAME

tinysh2 -- a tiny shell with a minimal set of features

=head1 SYNOPSIS

B<tinysh2>

=head1 DESCRIPTION

B<tinysh2> reads command lines from its standard input
and executes them. ``tinysh2> '' is given as prompt.

Each command line consists of space-separated tokens. Special tokens
begin with ``<'', indicating the input file, ``>'', indicating the
output file, ``>>'', specifying an output file that is to be extended,
or consist just of an ampersand ``&'', indicating a background execution. The
first non-special token specifies the command name which must be either
an absolute path of an executable file or locatable within the list of
directories provided by the environment variable B<PATH>.

=head1 DIAGNOSTICS

B<tinysh2> prints the exit code of a terminated program if it is non-zero.
An exit code of 255 is used by subprocesses if they are unable to start
the command for some reason.

=head1 BUGS

No pipelines, no builtins, no shell variables.

=head1 AUTHOR

Andreas F. Borchert

=cut

*/

#include <errno.h>
#include <signal.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <unistd.h>
#include <afblib/inbuf.h>
#include <afblib/inbuf_sareadline.h>
#include <afblib/outbuf.h>
#include <afblib/outbuf_printf.h>
#include <afblib/strlist.h>
#include <afblib/tokenizer.h>
#include "command.h"

static volatile sig_atomic_t interrupted = 0;

void interrupt_handler(int sig) {
   interrupted = 1;
}

inbuf in = {0};
outbuf out = {1};
outbuf err = {2};

/* print error much like perror but by using an outbuf */
void print_error(const char* s) {
   outbuf_printf(&err, "%s: %s\n", s, strerror(errno));
   outbuf_flush(&err);
   outbuf_free(&err);
}

void print_child_status(pid_t pid, int wstat) {
   outbuf_printf(&out, "[%d] ", (int) pid);
   if (WIFEXITED(wstat)) {
      outbuf_printf(&out, "exit %d", WEXITSTATUS(wstat));
   } else if (WIFSIGNALED(wstat)) {
      outbuf_printf(&out, "terminated with signal %d", WTERMSIG(wstat));
      if (WCOREDUMP(wstat)) outbuf_printf(&out, " (core dump)");
   } else if (WIFSTOPPED(wstat)) {
      outbuf_printf(&out, "stopped with signal %d", WSTOPSIG(wstat));
   } else if (WIFCONTINUED(wstat)) {
      outbuf_printf(&out, "continued");
   } else {
      outbuf_printf(&out, "???");
   }
   outbuf_printf(&out, "\n");
}

void status_report(void) {
   pid_t pid; int wstat;
   while ((pid = waitpid((pid_t)-1, &wstat, WNOHANG)) > 0) {
      print_child_status(pid, wstat);
   }
}

bool getline(stralloc* line) {
   bool first = true;
   interrupted = 0;
   for(;;) {
      if (interrupted) {
         interrupted = 0;
         outbuf_printf(&out, "\n");
         first = true;
      }
      if (first) {
         status_report();
         outbuf_printf(&out, "tinysh2> ");
         first = false;
      }
      errno = 0;
      outbuf_flush(&out);
      if (inbuf_sareadline(&in, line)) return true;
      if (errno != EINTR) return false;
   }
}

int main() {
   struct sigaction action = {
      .sa_handler = interrupt_handler,
   };
   if (sigaction(SIGINT, &action, 0) != 0 ||
         sigaction(SIGQUIT, &action, 0) != 0) {
      print_error("sigaction");
   }

   stralloc line = {0}; strlist tokens = {0}; command cmd = {0};
   while (getline(&line)) {
      stralloc_0(&line); /* required by tokenizer() */
      tokens.len = 0;
      if (!tokenizer(&line, &tokens)) break;
      if (tokens.len == 0) continue;
      if (!scan_command(&tokens, &cmd)) continue;

      sighold(SIGINT); sighold(SIGQUIT);
      pid_t child = fork();
      if (child == -1) {
         print_error("fork"); continue;
      }
      if (child == 0) {
         sigrelse(SIGINT); sigrelse(SIGQUIT);
         if (cmd.background) {
            sigignore(SIGINT); sigignore(SIGQUIT);
         }
         exec_command(&cmd);
	 print_error(cmd.cmdname);
         exit(255);
      }

      if (cmd.background) {
         outbuf_printf(&out, "%d\n", (int) child);
      } else {
         int wstat;
         pid_t pid = waitpid(child, &wstat, 0);
	 if (pid < 0) {
	    print_error("wait");
	 } else if (!WIFEXITED(wstat) || WEXITSTATUS(wstat)) {
            print_child_status(pid, wstat);
         }
      }
      sigrelse(SIGINT); sigrelse(SIGQUIT);
   }
} // main
